@@ -1,7 +1,7 @@
 #
 # phconvert - Reference library to read and save Photon-HDF5 files
 #
-# Copyright (C) 2014-2015 Antonino Ingargiola <tritemio@gmail.com>
+# Copyright (C) 2014-2017 Antonino Ingargiola <tritemio@gmail.com>
 #
 """
 
@@ -47,12 +47,24 @@ _EMPTY = ' '
 # Names of mandatory fields in the setup group
 _setup_mantatory_fields = ['num_pixels', 'num_spots', 'num_spectral_ch',
                            'num_polarization_ch', 'num_split_ch',
-                           'modulated_excitation', 'lifetime']
+                           'modulated_excitation', 'lifetime',
+                           'excitation_alternated']
 
 # Names of mandatory fields in the identity group
 _identity_mantatory_fields = ['format_name', 'format_version', 'format_url',
                               'software', 'software_version', 'creation_time']
 
+# Names of fields in /setup/detectors
+_detectors_group_fields = ('id', 'id_hardware', 'counts', 'dcr', 'afterpulsing',
+                           'positions', 'spot', 'module', 'label', 'tcspc_unit',
+                           'tcspc_num_bins')
+
+# All valid measurement_type strings
+valid_meas_types = ['smFRET', 'smFRET-usALEX', 'smFRET-usALEX-3c',
+                    'smFRET-nsALEX', 'generic']
+
+
+
 def _metapath(fullpath):
     """Normalize a HDF5 path by removing trailing digits after "photon_data".
     """
@@ -64,6 +76,7 @@ def _metapath(fullpath):
                     re.match(pattern, fullpath).group(1))
     return metapath
 
+
 def _analyze_path(name, prefix_list):
     """
     Analyze an HDF5 path.
@@ -106,12 +119,14 @@ def _analyze_path(name, prefix_list):
     return dict(full_path=full_path, group_path=group_path,
                 meta_path=meta_path, is_phdata=is_phdata, is_user=is_user)
 
+
 def _is_structured_array(obj):
     if hasattr(obj, 'dtype') and obj.dtype.kind == 'V':
         return True
     else:
         return False
 
+
 def _h5_write_array(group, name, obj, descr=None, chunked=False, h5file=None):
     """Writes `obj` in the pytables HDF5 `group` with name `name`.
     """
@@ -131,12 +146,16 @@ def _h5_write_array(group, name, obj, descr=None, chunked=False, h5file=None):
         if isinstance(obj, str):
             obj = obj.encode()
 
+    if name in h5file.get_node(group):
+        # Remove a pre-existing node with the same name
+        h5file.remove_node(group, name)
     save(group, name, obj=obj)
     # Set title through property access to work around pytable issue
     # under python 3 (https://github.com/PyTables/PyTables/issues/469)
     node = h5file.get_node(group)._f_get_child(name)
     node.title = descr.encode()  # saved as binary both on py2 and py3
 
+
 def _iter_hdf5_dict(data_dict, prefix_list=None, fields_descr=None,
                     debug=False):
     """Recursively iterate over `data_dict` returning a dict for each item.
@@ -172,6 +191,7 @@ def _iter_hdf5_dict(data_dict, prefix_list=None, fields_descr=None,
             if debug:
                 print('End Group "%s"' % (item['full_path']))
 
+
 def _save_photon_hdf5_dict(group, data_dict, fields_descr, prefix_list=None,
                            debug=False):
     """
@@ -193,16 +213,27 @@ def _save_photon_hdf5_dict(group, data_dict, fields_descr, prefix_list=None,
                 print('WARNING: missing description for "%s"' %
                       item['meta_path'])
 
-        if isinstance(item['value'], dict):
-            h5file.create_group(item['group_path'], item['name'],
-                                title=item['description'].encode())
+        if isinstance(item['value'], tables.Array):
+            # If the data is already a pytable array set only the title
+            item['value'].set_attr('TITLE', item['description'].encode())
+        elif isinstance(item['value'], dict):
+            if item['name'] in h5file.get_node(item['group_path']):
+                # If group exists only set TITLE
+                grp = h5file.get_node(item['group_path'], item['name'])
+                grp._f_setattr('TITLE', item['description'].encode())
+            else:
+                h5file.create_group(item['group_path'], item['name'],
+                                    title=item['description'].encode())
         else:
+            # Node is not a tables.Array or a Group, write it
             _h5_write_array(item['group_path'], item['name'],
                             obj=item['value'], descr=item['description'],
                             chunked=item['is_phdata'], h5file=group._v_file)
 
+
 def save_photon_hdf5(data_dict,
                      h5_fname = None,
+                     h5file = None,
                      user_descr = None,
                      overwrite = False,
                      compression = dict(complevel=6, complib='zlib'),
@@ -246,8 +277,15 @@ def save_photon_hdf5(data_dict,
     - `/setup/num_spots` (int): number of excitation/detection spots
     - `/setup/num_spectral_ch` (int): number of detection spectral bands
     - `/setup/num_polarization_ch` (int): number of detected polarization states
-    - `/setup/num_split_ch` (int): number of beam splitted channels
-    - `/setup/modulated_excitation` (bool): True if excitation is alternated.
+    - `/setup/num_split_ch` (int): number of beam split channels
+    - `/setup/modulated_excitation` (bool): True if there is any form of intensity
+      or polarization modulation or interleaved excitation (PIE or nsALEX).
+      This field has become obsolete in version 0.5 and maintained only for
+      compatibility.
+    - `/setup/excitation_alternated` (array of bool): New in version 0.5.
+      Values are True if the respective excitation source is
+      intensity-modulated. In us-ALEX both sources are alternated,
+      while in PAX measurements only one source is alternated.
     - `/setup/lifetime` (bool): True if dataset contains TCSPC data.
 
     See also
@@ -261,8 +299,13 @@ def save_photon_hdf5(data_dict,
             The keys must strings matching valid Photon-HDF5 paths.
             The values must be scalars, arrays, strings or another dict.
         h5_fname (string or None): file name for the output Photon-HDF5 file.
-            If None, the file name is taken from ``data_dict['_filename']``
-            with extension changed to '.hdf5'.
+            If None and h5file is also None, the file name is taken from
+            ``data_dict['_filename']`` with extension changed to '.hdf5'.
+        h5file (pytables.File or None): an already open and writable HDF5
+            file to use as container. This argument can be used to complete
+            an HDF5 file already containing some arrays, or to update
+            an already existing Photon-HDF5 file in-place.
+            For more info see note below.
         user_descr (dict or None): dictionary of descriptions (strings) for
             user-defined fields. The keys must be strings representing
             the full HDF5 path of each field. The values must be
@@ -289,31 +332,55 @@ def save_photon_hdf5(data_dict,
 
     For description and specs of the Photon-HDF5 format see:
     http://photon-hdf5.readthedocs.org/
+
+    Note:
+        The argument `h5file` accepts an already open HDF5 file for storage.
+        This allows completing a partially written file (for example
+        containing only photon_data arrays) or correcting and already complete
+        Photon-HDF5 file. When using `h5file`, you need to pass a full
+        `data_dict` structure as usual. If you don't want update an array,
+        put in `data_dict` a reference to the existing pytables array
+        (instead of using a numpy array). Fields containing numpy arrays
+        will be overwritten. Fields containing pytables Array (including
+        CArray or EArray) will be left unmodified. In either cases the TITLE
+        attribute is always updated.
     """
     comp_filter = tables.Filters(**compression)
 
     ## Compute file names
-    if h5_fname is None:
-        basename, extension = os.path.splitext(data_dict['_filename'])
-        if compression['complib'] == 'blosc':
-            basename += '_blosc'
-        h5_fname = basename + '.hdf5'
+    if h5file is not None:
+        _msg = 'Argument `h5file` must be None or a `tables.File` object.'
+        assert isinstance(h5file, tables.File), _msg
+        h5_fname = h5file.filename
+    else:
+        if h5_fname is None:
+            basename, extension = os.path.splitext(data_dict['_filename'])
+            if compression['complib'] == 'blosc':
+                basename += '_blosc'
+            h5_fname = basename + '.hdf5'
 
-    if os.path.isfile(h5_fname) and not overwrite:
-        basename, extension = os.path.splitext(h5_fname)
-        h5_fname = basename + '_new_copy.hdf5'
+        if os.path.isfile(h5_fname) and not overwrite:
+            basename, extension = os.path.splitext(h5_fname)
+            h5_fname = basename + '_new_copy.hdf5'
 
     ## Prefill and fix user-provided data_dict
     _populate_provenance(data_dict)
+    if 'setup' in data_dict:
+        _populate_setup(data_dict)
     _sanitize_data(data_dict, require_setup)
     _compute_acquisition_duration(data_dict)
 
     ## Create the HDF5 file
     print('Saving: %s' % h5_fname)
     title = official_fields_specs['/'][0].encode()
-    h5file = tables.open_file(h5_fname, mode="w", title=title,
-                              filters=comp_filter)
-    # Saving a file reference is useful in case of error
+    if h5file is None:
+        h5file = tables.open_file(str(h5_fname), mode="w", title=title,
+                                  filters=comp_filter)
+    else:
+        # If file already opened set only the root-node TITLE
+        h5file.set_node_attr('/', attrname='TITLE', attrvalue=title)
+
+    # Saving a file reference, useful in case of errors
     data_dict.update(_data_file=h5file)
 
     ## Identity info needs to be added after the file is created
@@ -339,6 +406,11 @@ def save_photon_hdf5(data_dict,
     if close:
         h5file.close()
 
+
+def _is_mutispot(h5root_or_dict):
+    return 'photon_data' not in h5root_or_dict
+
+
 def _populate_identity(data_dict, h5file):
     """Populate identity metadata adding info from the newly created file.
     """
@@ -349,6 +421,7 @@ def _populate_identity(data_dict, h5file):
         data_dict['identity'] = {}
     data_dict['identity'].update(identity)
 
+
 def _populate_provenance(data_dict):
     """Try to find the original data file to fill provenance fields.
     """
@@ -378,6 +451,7 @@ def _populate_provenance(data_dict):
         if orig_creation_time is not None:
             provenance['creation_time'] = orig_creation_time
 
+
 def _compute_acquisition_duration(data_dict):
     """Compute acquisition_duration if not present. Single-spot only.
     """
@@ -398,6 +472,56 @@ def _compute_acquisition_duration(data_dict):
                                 timestamps_unit)
         data_dict['acquisition_duration'] = np.round(acquisition_duration, 1)
 
+
+def _populate_setup(data_dict):
+    _populate_detectors_group(data_dict)
+    _autofill_laser_rep_rates(data_dict)
+
+
+def _autofill_laser_rep_rates(data_dict):
+    ph_data = data_dict[_sorted_photon_data(data_dict)[0]]
+    if 'measurement_specs' not in ph_data:
+        return
+    meas_specs = ph_data['measurement_specs']
+    if meas_specs['measurement_type'] != 'smFRET-nsALEX':
+        return
+    laser_rep_rate = 'laser_repetition_rate'
+    laser_rep_rates = laser_rep_rate + 's'
+    if laser_rep_rate not in meas_specs:
+        msg = """\
+            Measurement type 'smFRET-nsALEX' is missing the field
+            'laser_repetition_rate' in 'measurement_specs'."""
+        raise Invalid_PhotonHDF5(dedent(msg))
+    setup = data_dict['setup']
+    if laser_rep_rates not in setup:
+        # Create the array of laser repetition rates
+        # Note: all lasers repetition rates are assumed equal
+        setup[laser_rep_rates] = [meas_specs[laser_rep_rate]] * 2
+
+
+def _populate_detectors_group(data):
+    det_grp = data['setup'].get('detectors', {})
+    if 'id' in det_grp and 'id_hardware' in det_grp and 'counts' in det_grp:
+        # nothing to do
+        return
+
+    # Note that detectors with 0 counts cannot be detected
+    det_val, det_cnts, spot = [], [], []
+    for i, ph_data in enumerate(_sorted_photon_data(data)):
+        detectors = data[ph_data]['detectors'][:]  # numpy or pytables array
+        vals, counts = np.unique(detectors, return_counts=True)
+        det_val.extend(vals)
+        det_cnts.extend(counts)
+        spot.extend([i] * len(vals))
+
+    det_grp.setdefault('id', np.array(det_val))
+    det_grp.setdefault('id_hardware', np.array(det_val))
+    det_grp['counts'] = np.array(det_cnts)
+    if _is_mutispot(data):
+        det_grp['spot'] = np.array(spot)
+    data['setup']['detectors'] = det_grp
+
+
 def _get_identity(h5file):
     """Return a dict with identity information for `h5file`.
     """
@@ -412,6 +536,7 @@ def _get_identity(h5file):
                     format_url=root_attributes['format_url'])
     return identity
 
+
 def _get_file_metadata(fname):
     """Return a dict with file metadata.
     """
@@ -421,7 +546,7 @@ def _get_file_metadata(fname):
     filename = os.path.basename(full_filename)
 
     # Creation and modification time (but not exactly on *NIX)
-    # see https://docs.python.org/2/library/os.path.html#os.path.getctime)
+    # see https://docs.python.org/3/library/os.path.html#os.path.getctime)
     ctime = time.localtime(os.path.getctime(full_filename))
     mtime = time.localtime(os.path.getmtime(full_filename))
 
@@ -451,6 +576,7 @@ def dict_from_group(group, read=True):
         out[node._v_name] = value
     return out
 
+
 def dict_to_group(group, dictionary):
     """Save `dictionary` into HDF5 format in `group`.
     """
@@ -472,6 +598,7 @@ def dict_to_group(group, dictionary):
             node.title = _EMPTY.encode()  # saved as binary both on py2 and py3
     h5file.flush()
 
+
 def load_photon_hdf5(filename, **kwargs):
     """Open a Photon-HDF5 file in pytables, validating it.
 
@@ -517,6 +644,7 @@ def _get_version(h5file):
         raise Invalid_PhotonHDF5('No version identification.')
     return version
 
+
 def _check_version(filename):
     """Return file format version string (unicode on both py2 and py3).
 
@@ -528,6 +656,7 @@ def _check_version(filename):
         version = _get_version(h5file)
     return version
 
+
 def _sorted_photon_data_tables(h5file):
     """Return a sorted list of keys "photon_dataN", sorted by N.
 
@@ -538,9 +667,11 @@ def _sorted_photon_data_tables(h5file):
     ph_datas = [n for n in h5file.root._f_iter_nodes()
                 if n._v_name.startswith(prefix)]
 
-    ph_datas.sort(key=lambda x: x._v_name[len(prefix):])
+    if len(ph_datas) > 1:
+        ph_datas.sort(key=lambda x: int(x._v_name[len(prefix):]))
     return ph_datas
 
+
 def _sorted_photon_data(data_dict):
     """Return a sorted list of keys "photon_dataN", sorted by N.
 
@@ -550,10 +681,10 @@ def _sorted_photon_data(data_dict):
     prefix = 'photon_data'
     keys = [k for k in data_dict.keys() if k.startswith(prefix)]
     if len(keys) > 1:
-        sorted_channels = sorted([int(k[len(prefix):]) for k in keys])
-        keys = ['%s%d' % (prefix, ch) for ch in sorted_channels]
+        keys.sort(key=lambda x: int(x[len(prefix):]))
     return keys
 
+
 def photon_data_mapping(h5file, name='timestamps'):
     """Return a mapping (OrderedDict) between ch and photon_data array.
     """
@@ -567,6 +698,7 @@ def photon_data_mapping(h5file, name='timestamps'):
             mapping[ch] = ph
     return mapping
 
+
 def _is_sequence(obj):
     is_sequence = False
     if isinstance(obj, tuple) or isinstance(obj, list):
@@ -575,6 +707,7 @@ def _is_sequence(obj):
         is_sequence = obj.ndim > 0
     return is_sequence
 
+
 def _normalize_bools(data_dict):
     """Cast bools (both scalars or in sequences) to integers."""
     for name, value in data_dict.items():
@@ -586,6 +719,7 @@ def _normalize_bools(data_dict):
             elif _is_sequence(value) and isinstance(value[0], bool):
                 data_dict[name] = np.asarray(value, dtype='uint8')
 
+
 def _normalize_detectors_specs(data_dict):
     base = '/photon_data/measurement_specs/detectors_specs/'
     names = ['spectral_ch1', 'spectral_ch2', 'split_ch1', 'split_ch2',
@@ -594,12 +728,16 @@ def _normalize_detectors_specs(data_dict):
 
     # Retrive the detectors' dtype (from first photon_data group in multi-spot)
     ph_data = data_dict[_sorted_photon_data(data_dict)[0]]
+    if 'detectors' not in ph_data:
+        return
     dtype = ph_data['detectors'].dtype
 
     for item in _iter_hdf5_dict(data_dict):
         if item['meta_path'] in cast_fields:
             cdict = item['curr_dict']
-            cdict[item['name']] = np.array(item['value'], dtype=dtype, ndmin=1)
+            cdict[item['name']] = np.array(item['value'],
+                                           dtype=dtype, ndmin=1)
+
 
 def _normalize_setup_arrays(data_dict):
     """Make sure arrays of float in setup are arrays of floats."""
@@ -611,10 +749,23 @@ def _normalize_setup_arrays(data_dict):
     # Arrays of float fields in setup group
     names_aof = ['detection_wavelengths', 'excitation_wavelengths',
                  'excitation_input_powers', 'detection_polarizations',
-                 'excitation_intensity', 'detection_split_ch_ratios']
+                 'excitation_intensity', 'detection_split_ch_ratios',
+                 'laser_repetition_rates']
     for name in names_aof:
         if name in setup:
-            setup[name] = np.array([float(v) for v in setup[name]], dtype=float)
+            setup[name] = np.array([float(v) for v in setup[name]],
+                                   dtype=float)
+
+
+def _normalize_detectors_group(data_dict):
+    """Convert fields in /setup/detectors to `numpy.ndarray`."""
+    if 'setup' not in data_dict:
+        return
+    det_grp = data_dict['setup']['detectors']
+    for name in _detectors_group_fields:
+        if name in det_grp:
+            det_grp[name] = np.asarray(det_grp[name])
+
 
 def _convert_scalar_item(item):
     """Cast a scalar item (from _iter_hdf5_dict) to scalar."""
@@ -647,9 +798,10 @@ def _convert_scalar_item(item):
                                      % item['meta_path'])
     return scalar_value
 
+
 def _normalize_scalars(data_dict):
     """Make sure all scalar fields are scalars."""
-    ## scalar fields conversions
+    # scalar fields conversions
     for item in _iter_hdf5_dict(data_dict):
         if item['is_user']:
             continue
@@ -658,6 +810,7 @@ def _normalize_scalars(data_dict):
             curr_dict = item['curr_dict']
             curr_dict[item['name']] = scalar_value
 
+
 def _sanitize_data(data_dict, require_setup=True):
     """Perform type conversions to strictly conform to Photon-HDF5 specs.
 
@@ -667,7 +820,8 @@ def _sanitize_data(data_dict, require_setup=True):
     - convert scalar fields that are array of size == 1 to scalars
     - cast bools or sequences of bools to integers
     - convert scalar fields which are strings to numbers
-    - convert sequences of strings in arrays of floats for selected setup fields
+    - for some setup fields, convert sequences of strings to arrays of floats
+    - convert /setup/detectors fields into numpy arrays.
     """
     def _assert_has_key(dict_, key, dict_name):
         if key not in dict_:
@@ -697,15 +851,20 @@ def _sanitize_data(data_dict, require_setup=True):
     _normalize_setup_arrays(data_dict)
     # Cast scalar fields to scalar
     _normalize_scalars(data_dict)
+    # Convert fields in /setup/detectors to numpy arrays
+    _normalize_detectors_group(data_dict)
+
 
 ##
 # Validation functions
 #
+
 class Invalid_PhotonHDF5(Exception):
     """Error raised when a file is not a valid Photon-HDF5 file.
     """
     pass
 
+
 def _assert_valid(condition, msg, strict=True, norepeat=False, pool=None):
     """Assert `condition` and raise Invalid_PhotonHDF5(msg) on fail.
 
@@ -736,6 +895,7 @@ def _assert_valid(condition, msg, strict=True, norepeat=False, pool=None):
             print('Photon-HDF5 WARNING: %s' % msg)
     return condition
 
+
 def _assert_has_field(name, group, msg=None, msg_add=None, mandatory=True,
                       norepeat=False, pool=None, verbose=False):
     """Assert that field `name` is in `group`.
@@ -761,12 +921,43 @@ def _assert_has_field(name, group, msg=None, msg_add=None, mandatory=True,
     if verbose:
         print('Checking "%s" in %s.' % (name, group._v_pathname))
     if msg is None:
-        msg = 'Missing field "%s" in "%s".' % (name, group._v_pathname)
+        msg = 'Missing field "%s" in "%s". ' % (name, group._v_pathname)
     if msg_add is not None:
         msg += msg_add
     return _assert_valid(name in group, msg, mandatory, norepeat, pool)
 
 
+def _assert_valid_detectors(h5file):
+    detectors = h5file.root.setup.detectors
+    det_ids = detectors.id.read()
+    if 'counts' in detectors:
+        det_counts = detectors.counts.read()
+    if _is_mutispot(h5file.root):
+        spot = detectors.spot.read()
+    else:
+        spot = np.zeros(len(det_ids), dtype='uint8')
+
+    m = 'detectors/%s length (%d) is not equal to the number of detectors (%d).'
+    for field in _detectors_group_fields:
+        if field in detectors:
+            values = detectors._f_get_child(field)
+            _assert_valid(len(values) == len(det_ids),
+                          msg=m % (field, len(values), len(det_ids)))
+
+    msg = 'Detector %d in spot %d not found in detectors/id.'
+    msgc = 'Wrong counts (%d instead of %d) for detector %d in spot %d.'
+    for i, ph_data in enumerate(_sorted_photon_data_tables(h5file)):
+        if 'detectors' not in ph_data:
+            break
+        vals, cnts = np.unique(ph_data.detectors[:], return_counts=True)
+        det_ids_spot = det_ids[spot == i]
+        for v, c in zip(vals, cnts):
+            _assert_valid(v in det_ids_spot, msg=msg % (v, i))
+            if 'counts' in detectors:
+                csaved = det_counts[spot == i][det_ids_spot == v]
+                _assert_valid(c == csaved, msg=msgc % (csaved, c, v, i))
+
+
 def assert_valid_photon_hdf5(datafile, warnings=True, verbose=False,
                              strict_description=True, require_setup=True,
                              skip_measurement_specs=False):
@@ -818,32 +1009,30 @@ def assert_valid_photon_hdf5(datafile, warnings=True, verbose=False,
     _assert_identity(h5file, warnings=warnings, verbose=verbose)
 
     pool = []
-    kwargs = dict(pool=pool, norepeat=True,
+    kwargs = dict(setup=h5file.root.setup, pool=pool, norepeat=True,
                   skip_measurement_specs=skip_measurement_specs)
     for ph_data in _sorted_photon_data_tables(h5file):
         _check_photon_data_tables(ph_data, **kwargs)
-        if '/setup/lifetime' in h5file and h5file.root.setup.lifetime.read():
-            _assert_has_field('nanotimes', ph_data, verbose=verbose)
-            _assert_has_field('nanotimes_specs', ph_data, verbose=verbose)
-            nt_specs = ph_data.nanotimes_specs
-            _assert_has_field('tcspc_unit', nt_specs, verbose=verbose)
-            _assert_has_field('tcspc_num_bins', nt_specs, verbose=verbose)
+
 
 def _assert_setup(h5file, warnings=True, strict=True, verbose=False):
     """Assert that setup exists and contains the mandatory fields.
     """
-    if _assert_has_field('setup', h5file.root, mandatory=strict,
-                         verbose=verbose):
-        for name in _setup_mantatory_fields:
-            _assert_has_field(name, h5file.root.setup, mandatory=strict,
-                              verbose=verbose)
-        if not warnings:
-            return
+    if not _assert_has_field('setup', h5file.root, mandatory=strict,
+                             verbose=verbose):
+        return
+    for name in _setup_mantatory_fields:
+        _assert_has_field(name, h5file.root.setup, mandatory=strict,
+                          verbose=verbose)
+    if 'detectors' in h5file.root.setup:
+        _assert_valid_detectors(h5file)
+    if warnings:
         optional_fields = ['excitation_wavelengths', 'detection_wavelengths']
         for name in optional_fields:
             _assert_has_field(name, h5file.root.setup, mandatory=False,
                               verbose=verbose)
 
+
 def _assert_identity(h5file, warnings=True, strict=True, verbose=False):
     """Assert that identity group exists and contains the mandatory fields.
     """
@@ -859,6 +1048,7 @@ def _assert_identity(h5file, warnings=True, strict=True, verbose=False):
             _assert_has_field(name, h5file.root.identity, mandatory=False,
                               verbose=verbose)
 
+
 def _assert_valid_fields(h5file, strict_description=True, verbose=False):
     """Assert compliance of field names, descriptions and data types.
 
@@ -877,7 +1067,7 @@ def _assert_valid_fields(h5file, strict_description=True, verbose=False):
         _assert_valid(len(title) > 0, msg, strict=strict_description)
 
         ## Test description is a binary string
-        # This depends on how pytbales loads the string and fails for some
+        # This depends on how pytables loads the string and fails for some
         # fields (e.g. user fields in BH file) under python 3.
         # The test is disable for the time being.
         #msg = 'TITLE attribute for "%s" is not a binary string.' % pathname
@@ -915,7 +1105,8 @@ def _assert_valid_fields(h5file, strict_description=True, verbose=False):
             else:
                 raise ValueError('Wrong type in JSON specs.')
 
-def _check_photon_data_tables(ph_data, norepeat=False, pool=None,
+
+def _check_photon_data_tables(ph_data, setup, norepeat=False, pool=None,
                               skip_measurement_specs=False, verbose=False):
     """Assert that the photon_data group follows the Photon-HDF5 specs.
     """
@@ -931,8 +1122,6 @@ def _check_photon_data_tables(ph_data, norepeat=False, pool=None,
                               verbose=verbose, norepeat=norepeat, pool=pool)
         return
 
-    spectral_meas_types = ['smFRET', 'smFRET-usALEX', 'smFRET-usALEX-3c',
-                           'smFRET-nsALEX']
     meas_specs = ph_data.measurement_specs
     msg = 'Missing "measurement_type" in "%s".' % meas_specs._v_pathname
     _assert_has_field('measurement_type', meas_specs, msg, verbose=verbose)
@@ -940,25 +1129,97 @@ def _check_photon_data_tables(ph_data, norepeat=False, pool=None,
     meas_type = meas_specs.measurement_type.read().decode()
     if verbose:
         print('* Measurement type: "%s"' % meas_type)
-    _assert_valid(meas_type in spectral_meas_types,
-                  msg='Unkwnown measurement type "%s"' % meas_type)
+    _assert_valid(meas_type in valid_meas_types,
+                  msg='Unknown measurement type "%s"' % meas_type)
 
     # At this point we have a valid measurement_type
-    # Any missing field will raise an error.
+    # We will check (and raise an error) for any missing field.
     msg = '\nThis field is mandatory for "%s" data.' % meas_type
     kwargs = dict(msg_add=msg, verbose=verbose)
-    _assert_has_field('spectral_ch1', meas_specs.detectors_specs, **kwargs)
-    _assert_has_field('spectral_ch2', meas_specs.detectors_specs, **kwargs)
 
-    if meas_type in ['smFRET-usALEX', 'smFRET-usALEX-3c']:
-        _assert_has_field('alex_period', meas_specs, **kwargs)
+    # Read number of channels in each branch
+    num_ch = dict(spectral=setup.num_spectral_ch.read(),
+                  split=setup.num_split_ch.read(),
+                  polarization=setup.num_polarization_ch.read())
 
-    if meas_type == 'smFRET-nsALEX':
-        _assert_has_field('laser_repetition_rate', meas_specs, **kwargs)
+    # Check for spectral channels
+    if meas_type in ('smFRET', 'smFRET-usALEX', 'smFRET-nsALEX'):
+        _msg = ('%s measurement requires /setup/num_spectral_ch = 2 '
+                '(not %d).' % (meas_type, num_ch['spectral']))
+        _assert_valid(num_ch['spectral'] == 2, msg=_msg)
+    if meas_type == 'smFRET-usALEX-3c':
+        _msg = ('%s measurement requires /setup/num_spectral_ch = 3 '
+                '(not %d).' % (meas_type, num_ch['spectral']))
+        _assert_valid(num_ch['spectral'] == 3, msg=_msg)
+
+    # handle case of no detectors array and no detectors_specs
+    # which is possible when measurement_type is 'generic'
+    if 'detectors' in ph_data:
+        det_specs = meas_specs.detectors_specs
+        # Check for spectral/split/polarization channels in detectors_specs
+        for feature, nch in num_ch.items():
+            if nch > 1:
+                for i in range(nch):
+                    _assert_has_field('%s_ch%d' % (feature, i + 1),
+                                      det_specs, **kwargs)
+
+    # Check presences of repetititon rate with pulsed lasers
+    msg0 = "According to /setup/excitation_cw some lasers are pulsed.\n"
+    if not all(setup.excitation_cw[:]):
+        m = "However, the field '/setup/laser_repetition_rates' is missing."
+        _assert_has_field('laser_repetition_rates', setup, msg_add=msg0 + m)
+        m = ("However, the field 'measurement_specs/laser_repetition_rate' "
+             "is missing.")
+        _assert_has_field('laser_repetition_rate', meas_specs,
+                          msg_add=msg0 + m)
+
+    msg_cw = """
+    According to /setup/excitation_alternated this measurement uses
+    laser alternation and CW lasers. However, there is no alex_period
+    field in measurement_specs. alex_period is mandatory in measurements
+    using alternation and CW lasers."""
+    if all(setup.excitation_cw[:]):
+        if any(setup.excitation_alternated[:]):
+            _assert_has_field('alex_period', meas_specs,
+                              msg_add=dedent(msg_cw))
+
+    if 'nanotimes' in ph_data and 'lifetime' in setup:
+        _assert_valid(setup.lifetime.read(),
+                      'Lifetime is False but nanotimes are present.')
+
+    if 'lifetime' in setup and setup.lifetime.read():
+        msg = """\
+        According to /setup/lifetime (=True) this file should be a
+        TCSPC measurement. However /setup/excitation_cw says that all
+        the laser sources are CW instead of pulsed. At least one source
+        needs to be pulsed."""
+        _assert_valid(not all(setup.excitation_cw.read()), msg=dedent(msg))
         _assert_has_field('nanotimes', ph_data, **kwargs)
-        _assert_has_field('nanotimes_specs', ph_data, **kwargs)
-        for name in ['tcspc_unit', 'tcspc_num_bins']:
-            _assert_has_field(name, ph_data.nanotimes_specs, **kwargs)
+
+        if 'nanotimes_specs' in ph_data:
+            _assert_has_field('nanotimes_specs', ph_data, **kwargs)
+            tcspc_specs_group = ph_data.nanotimes_specs
+        else:
+            _assert_has_field('detectors', setup, **kwargs)
+            tcspc_specs_group = setup.detectors
+        for name in ('tcspc_unit', 'tcspc_num_bins'):
+            _assert_has_field(name, tcspc_specs_group, **kwargs)
+
+    # us-ALEX fields
+    if meas_type in ('smFRET-usALEX', 'smFRET-usALEX-3c'):
+        msg = 'All lasers need to be CW in %s measurements.'
+        _assert_valid(all(setup.excitation_cw[:]), msg=msg % meas_type)
+        msg = 'All lasers need to be alternated in %s measurements.'
+        _assert_valid(all(setup.excitation_alternated[:]),
+                      msg=msg % meas_type)
+
+    # ns-ALEX / PIE fields
+    if meas_type == 'smFRET-nsALEX':
+        msg = 'All lasers need to be pulsed in smFRET-nsALEX measurements.'
+        _assert_valid(all(~setup.excitation_cw[:]), msg=msg)
+        _assert_has_field('lifetime', setup, **kwargs)
+        _assert_valid(setup.lifetime.read(),
+                      msg='smFRET-nsALEX requires lifetime = True.')
 
 
 def print_attrs(node, which='user'):
